# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf/remote_call.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import streaming_pb2 as protobuf_dot_streaming__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf/remote_call.proto',
  package='ray.streaming.proto',
  syntax='proto3',
  serialized_options=_b('\n\"io.ray.streaming.runtime.generated'),
  serialized_pb=_b('\n\x1aprotobuf/remote_call.proto\x12\x13ray.streaming.proto\x1a\x18protobuf/streaming.proto\x1a\x19google/protobuf/any.proto\"\xbb\x0c\n\x16\x45xecutionVertexContext\x12u\n\x18\x63urrent_execution_vertex\x18\x01 \x01(\x0b\x32;.ray.streaming.proto.ExecutionVertexContext.ExecutionVertexR\x16\x63urrentExecutionVertex\x12{\n\x1bupstream_execution_vertices\x18\x02 \x03(\x0b\x32;.ray.streaming.proto.ExecutionVertexContext.ExecutionVertexR\x19upstreamExecutionVertices\x12\x7f\n\x1d\x64ownstream_execution_vertices\x18\x03 \x03(\x0b\x32;.ray.streaming.proto.ExecutionVertexContext.ExecutionVertexR\x1b\x64ownstreamExecutionVertices\x12m\n\x15input_execution_edges\x18\x04 \x03(\x0b\x32\x39.ray.streaming.proto.ExecutionVertexContext.ExecutionEdgeR\x13inputExecutionEdges\x12o\n\x16output_execution_edges\x18\x05 \x03(\x0b\x32\x39.ray.streaming.proto.ExecutionVertexContext.ExecutionEdgeR\x14outputExecutionEdges\x1a\xa7\x01\n\rExecutionEdge\x12;\n\x1asource_execution_vertex_id\x18\x01 \x01(\x05R\x17sourceExecutionVertexId\x12;\n\x1atarget_execution_vertex_id\x18\x02 \x01(\x05R\x17targetExecutionVertexId\x12\x1c\n\tpartition\x18\x03 \x01(\x0cR\tpartition\x1a\xa1\x06\n\x0f\x45xecutionVertex\x12.\n\x13\x65xecution_vertex_id\x18\x01 \x01(\x05R\x11\x65xecutionVertexId\x12\x35\n\x17\x65xecution_job_vertex_id\x18\x02 \x01(\x05R\x14\x65xecutionJobVertexId\x12\x39\n\x19\x65xecution_job_vertex_name\x18\x03 \x01(\tR\x16\x65xecutionJobVertexName\x12\x34\n\x16\x65xecution_vertex_index\x18\x04 \x01(\x05R\x14\x65xecutionVertexIndex\x12 \n\x0bparallelism\x18\x05 \x01(\x05R\x0bparallelism\x12\x1a\n\x08operator\x18\x06 \x01(\x0cR\x08operator\x12\x18\n\x07\x63hained\x18\x07 \x01(\x08R\x07\x63hained\x12!\n\x0cworker_actor\x18\x08 \x01(\x0cR\x0bworkerActor\x12!\n\x0c\x63ontainer_id\x18\t \x01(\tR\x0b\x63ontainerId\x12\x1d\n\nbuild_time\x18\n \x01(\x04R\tbuildTime\x12\x39\n\x08language\x18\x0b \x01(\x0e\x32\x1d.ray.streaming.proto.LanguageR\x08language\x12_\n\x06\x63onfig\x18\x0c \x03(\x0b\x32G.ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ConfigEntryR\x06\x63onfig\x12\x65\n\x08resource\x18\r \x03(\x0b\x32I.ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ResourceEntryR\x08resource\x1a\x39\n\x0b\x43onfigEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a;\n\rResourceEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xa2\x01\n\x16PythonJobWorkerContext\x12!\n\x0cmaster_actor\x18\x01 \x01(\x0cR\x0bmasterActor\x12\x65\n\x18\x65xecution_vertex_context\x18\x02 \x01(\x0b\x32+.ray.streaming.proto.ExecutionVertexContextR\x16\x65xecutionVertexContext\"&\n\nBoolResult\x12\x18\n\x07\x62oolRes\x18\x01 \x01(\x08R\x07\x62oolRes\"\x19\n\x07\x42\x61rrier\x12\x0e\n\x02id\x18\x01 \x01(\x03R\x02id\"3\n\x0c\x43heckpointId\x12#\n\rcheckpoint_id\x18\x01 \x01(\x03R\x0c\x63heckpointId\"v\n\rBaseWorkerCmd\x12\x19\n\x08\x61\x63tor_id\x18\x01 \x01(\x0cR\x07\x61\x63torId\x12\x1c\n\ttimestamp\x18\x02 \x01(\x03R\ttimestamp\x12,\n\x06\x64\x65tail\x18\x03 \x01(\x0b\x32\x14.google.protobuf.AnyR\x06\x64\x65tail\"F\n\x12WorkerCommitReport\x12\x30\n\x14\x63ommit_checkpoint_id\x18\x01 \x01(\x03R\x12\x63ommitCheckpointId\"\x84\x01\n\x15WorkerRollbackRequest\x12#\n\rexception_msg\x18\x01 \x01(\tR\x0c\x65xceptionMsg\x12\'\n\x0fworker_hostname\x18\x02 \x01(\tR\x0eworkerHostname\x12\x1d\n\nworker_pid\x18\x03 \x01(\tR\tworkerPid\"\xac\x01\n\nCallResult\x12\x18\n\x07success\x18\x01 \x01(\x08R\x07success\x12\x1f\n\x0bresult_code\x18\x02 \x01(\x05R\nresultCode\x12\x1d\n\nresult_msg\x18\x03 \x01(\tR\tresultMsg\x12\x44\n\nresult_obj\x18\x04 \x01(\x0b\x32%.ray.streaming.proto.QueueRecoverInfoR\tresultObj\"\xc4\x02\n\x10QueueRecoverInfo\x12\x62\n\x0f\x63reation_status\x18\x03 \x03(\x0b\x32\x39.ray.streaming.proto.QueueRecoverInfo.CreationStatusEntryR\x0e\x63reationStatus\x1a|\n\x13\x43reationStatusEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12O\n\x05value\x18\x02 \x01(\x0e\x32\x39.ray.streaming.proto.QueueRecoverInfo.QueueCreationStatusR\x05value:\x02\x38\x01\"N\n\x13QueueCreationStatus\x12\x10\n\x0c\x46reshStarted\x10\x00\x12\n\n\x06PullOk\x10\x01\x12\x0b\n\x07Timeout\x10\x02\x12\x0c\n\x08\x44\x61taLost\x10\x03\x42$\n\"io.ray.streaming.runtime.generatedb\x06proto3')
  ,
  dependencies=[protobuf_dot_streaming__pb2.DESCRIPTOR,google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])



_QUEUERECOVERINFO_QUEUECREATIONSTATUS = _descriptor.EnumDescriptor(
  name='QueueCreationStatus',
  full_name='ray.streaming.proto.QueueRecoverInfo.QueueCreationStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FreshStarted', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PullOk', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Timeout', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DataLost', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2736,
  serialized_end=2814,
)
_sym_db.RegisterEnumDescriptor(_QUEUERECOVERINFO_QUEUECREATIONSTATUS)


_EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE = _descriptor.Descriptor(
  name='ExecutionEdge',
  full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionEdge',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_execution_vertex_id', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionEdge.source_execution_vertex_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceExecutionVertexId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target_execution_vertex_id', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionEdge.target_execution_vertex_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='targetExecutionVertexId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='partition', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionEdge.partition', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='partition', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=729,
  serialized_end=896,
)

_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY = _descriptor.Descriptor(
  name='ConfigEntry',
  full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ConfigEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ConfigEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ConfigEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1582,
  serialized_end=1639,
)

_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY = _descriptor.Descriptor(
  name='ResourceEntry',
  full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ResourceEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ResourceEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ResourceEntry.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1641,
  serialized_end=1700,
)

_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX = _descriptor.Descriptor(
  name='ExecutionVertex',
  full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='execution_vertex_id', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.execution_vertex_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='executionVertexId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='execution_job_vertex_id', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.execution_job_vertex_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='executionJobVertexId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='execution_job_vertex_name', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.execution_job_vertex_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='executionJobVertexName', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='execution_vertex_index', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.execution_vertex_index', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='executionVertexIndex', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parallelism', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.parallelism', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='parallelism', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operator', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.operator', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='operator', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chained', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.chained', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='chained', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='worker_actor', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.worker_actor', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerActor', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='container_id', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.container_id', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='containerId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='build_time', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.build_time', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='buildTime', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='language', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.language', index=10,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='language', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='config', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.config', index=11,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='config', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='resource', full_name='ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.resource', index=12,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='resource', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY, _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=899,
  serialized_end=1700,
)

_EXECUTIONVERTEXCONTEXT = _descriptor.Descriptor(
  name='ExecutionVertexContext',
  full_name='ray.streaming.proto.ExecutionVertexContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='current_execution_vertex', full_name='ray.streaming.proto.ExecutionVertexContext.current_execution_vertex', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='currentExecutionVertex', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='upstream_execution_vertices', full_name='ray.streaming.proto.ExecutionVertexContext.upstream_execution_vertices', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='upstreamExecutionVertices', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='downstream_execution_vertices', full_name='ray.streaming.proto.ExecutionVertexContext.downstream_execution_vertices', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='downstreamExecutionVertices', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='input_execution_edges', full_name='ray.streaming.proto.ExecutionVertexContext.input_execution_edges', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='inputExecutionEdges', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_execution_edges', full_name='ray.streaming.proto.ExecutionVertexContext.output_execution_edges', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='outputExecutionEdges', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE, _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=1700,
)


_PYTHONJOBWORKERCONTEXT = _descriptor.Descriptor(
  name='PythonJobWorkerContext',
  full_name='ray.streaming.proto.PythonJobWorkerContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_actor', full_name='ray.streaming.proto.PythonJobWorkerContext.master_actor', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='masterActor', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='execution_vertex_context', full_name='ray.streaming.proto.PythonJobWorkerContext.execution_vertex_context', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='executionVertexContext', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1703,
  serialized_end=1865,
)


_BOOLRESULT = _descriptor.Descriptor(
  name='BoolResult',
  full_name='ray.streaming.proto.BoolResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='boolRes', full_name='ray.streaming.proto.BoolResult.boolRes', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='boolRes', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1867,
  serialized_end=1905,
)


_BARRIER = _descriptor.Descriptor(
  name='Barrier',
  full_name='ray.streaming.proto.Barrier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ray.streaming.proto.Barrier.id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1907,
  serialized_end=1932,
)


_CHECKPOINTID = _descriptor.Descriptor(
  name='CheckpointId',
  full_name='ray.streaming.proto.CheckpointId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='checkpoint_id', full_name='ray.streaming.proto.CheckpointId.checkpoint_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='checkpointId', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1934,
  serialized_end=1985,
)


_BASEWORKERCMD = _descriptor.Descriptor(
  name='BaseWorkerCmd',
  full_name='ray.streaming.proto.BaseWorkerCmd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='actor_id', full_name='ray.streaming.proto.BaseWorkerCmd.actor_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='actorId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='ray.streaming.proto.BaseWorkerCmd.timestamp', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timestamp', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detail', full_name='ray.streaming.proto.BaseWorkerCmd.detail', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='detail', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1987,
  serialized_end=2105,
)


_WORKERCOMMITREPORT = _descriptor.Descriptor(
  name='WorkerCommitReport',
  full_name='ray.streaming.proto.WorkerCommitReport',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit_checkpoint_id', full_name='ray.streaming.proto.WorkerCommitReport.commit_checkpoint_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='commitCheckpointId', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2107,
  serialized_end=2177,
)


_WORKERROLLBACKREQUEST = _descriptor.Descriptor(
  name='WorkerRollbackRequest',
  full_name='ray.streaming.proto.WorkerRollbackRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exception_msg', full_name='ray.streaming.proto.WorkerRollbackRequest.exception_msg', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='exceptionMsg', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='worker_hostname', full_name='ray.streaming.proto.WorkerRollbackRequest.worker_hostname', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerHostname', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='worker_pid', full_name='ray.streaming.proto.WorkerRollbackRequest.worker_pid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerPid', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2180,
  serialized_end=2312,
)


_CALLRESULT = _descriptor.Descriptor(
  name='CallResult',
  full_name='ray.streaming.proto.CallResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='ray.streaming.proto.CallResult.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='success', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_code', full_name='ray.streaming.proto.CallResult.result_code', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='resultCode', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_msg', full_name='ray.streaming.proto.CallResult.result_msg', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='resultMsg', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_obj', full_name='ray.streaming.proto.CallResult.result_obj', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='resultObj', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2315,
  serialized_end=2487,
)


_QUEUERECOVERINFO_CREATIONSTATUSENTRY = _descriptor.Descriptor(
  name='CreationStatusEntry',
  full_name='ray.streaming.proto.QueueRecoverInfo.CreationStatusEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ray.streaming.proto.QueueRecoverInfo.CreationStatusEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ray.streaming.proto.QueueRecoverInfo.CreationStatusEntry.value', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2610,
  serialized_end=2734,
)

_QUEUERECOVERINFO = _descriptor.Descriptor(
  name='QueueRecoverInfo',
  full_name='ray.streaming.proto.QueueRecoverInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='creation_status', full_name='ray.streaming.proto.QueueRecoverInfo.creation_status', index=0,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='creationStatus', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_QUEUERECOVERINFO_CREATIONSTATUSENTRY, ],
  enum_types=[
    _QUEUERECOVERINFO_QUEUECREATIONSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2490,
  serialized_end=2814,
)

_EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE.containing_type = _EXECUTIONVERTEXCONTEXT
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY.containing_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY.containing_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX.fields_by_name['language'].enum_type = protobuf_dot_streaming__pb2._LANGUAGE
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX.fields_by_name['config'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX.fields_by_name['resource'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX.containing_type = _EXECUTIONVERTEXCONTEXT
_EXECUTIONVERTEXCONTEXT.fields_by_name['current_execution_vertex'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX
_EXECUTIONVERTEXCONTEXT.fields_by_name['upstream_execution_vertices'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX
_EXECUTIONVERTEXCONTEXT.fields_by_name['downstream_execution_vertices'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX
_EXECUTIONVERTEXCONTEXT.fields_by_name['input_execution_edges'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE
_EXECUTIONVERTEXCONTEXT.fields_by_name['output_execution_edges'].message_type = _EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE
_PYTHONJOBWORKERCONTEXT.fields_by_name['execution_vertex_context'].message_type = _EXECUTIONVERTEXCONTEXT
_BASEWORKERCMD.fields_by_name['detail'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_CALLRESULT.fields_by_name['result_obj'].message_type = _QUEUERECOVERINFO
_QUEUERECOVERINFO_CREATIONSTATUSENTRY.fields_by_name['value'].enum_type = _QUEUERECOVERINFO_QUEUECREATIONSTATUS
_QUEUERECOVERINFO_CREATIONSTATUSENTRY.containing_type = _QUEUERECOVERINFO
_QUEUERECOVERINFO.fields_by_name['creation_status'].message_type = _QUEUERECOVERINFO_CREATIONSTATUSENTRY
_QUEUERECOVERINFO_QUEUECREATIONSTATUS.containing_type = _QUEUERECOVERINFO
DESCRIPTOR.message_types_by_name['ExecutionVertexContext'] = _EXECUTIONVERTEXCONTEXT
DESCRIPTOR.message_types_by_name['PythonJobWorkerContext'] = _PYTHONJOBWORKERCONTEXT
DESCRIPTOR.message_types_by_name['BoolResult'] = _BOOLRESULT
DESCRIPTOR.message_types_by_name['Barrier'] = _BARRIER
DESCRIPTOR.message_types_by_name['CheckpointId'] = _CHECKPOINTID
DESCRIPTOR.message_types_by_name['BaseWorkerCmd'] = _BASEWORKERCMD
DESCRIPTOR.message_types_by_name['WorkerCommitReport'] = _WORKERCOMMITREPORT
DESCRIPTOR.message_types_by_name['WorkerRollbackRequest'] = _WORKERROLLBACKREQUEST
DESCRIPTOR.message_types_by_name['CallResult'] = _CALLRESULT
DESCRIPTOR.message_types_by_name['QueueRecoverInfo'] = _QUEUERECOVERINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExecutionVertexContext = _reflection.GeneratedProtocolMessageType('ExecutionVertexContext', (_message.Message,), {

  'ExecutionEdge' : _reflection.GeneratedProtocolMessageType('ExecutionEdge', (_message.Message,), {
    'DESCRIPTOR' : _EXECUTIONVERTEXCONTEXT_EXECUTIONEDGE,
    '__module__' : 'protobuf.remote_call_pb2'
    # @@protoc_insertion_point(class_scope:ray.streaming.proto.ExecutionVertexContext.ExecutionEdge)
    })
  ,

  'ExecutionVertex' : _reflection.GeneratedProtocolMessageType('ExecutionVertex', (_message.Message,), {

    'ConfigEntry' : _reflection.GeneratedProtocolMessageType('ConfigEntry', (_message.Message,), {
      'DESCRIPTOR' : _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY,
      '__module__' : 'protobuf.remote_call_pb2'
      # @@protoc_insertion_point(class_scope:ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ConfigEntry)
      })
    ,

    'ResourceEntry' : _reflection.GeneratedProtocolMessageType('ResourceEntry', (_message.Message,), {
      'DESCRIPTOR' : _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY,
      '__module__' : 'protobuf.remote_call_pb2'
      # @@protoc_insertion_point(class_scope:ray.streaming.proto.ExecutionVertexContext.ExecutionVertex.ResourceEntry)
      })
    ,
    'DESCRIPTOR' : _EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX,
    '__module__' : 'protobuf.remote_call_pb2'
    # @@protoc_insertion_point(class_scope:ray.streaming.proto.ExecutionVertexContext.ExecutionVertex)
    })
  ,
  'DESCRIPTOR' : _EXECUTIONVERTEXCONTEXT,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.ExecutionVertexContext)
  })
_sym_db.RegisterMessage(ExecutionVertexContext)
_sym_db.RegisterMessage(ExecutionVertexContext.ExecutionEdge)
_sym_db.RegisterMessage(ExecutionVertexContext.ExecutionVertex)
_sym_db.RegisterMessage(ExecutionVertexContext.ExecutionVertex.ConfigEntry)
_sym_db.RegisterMessage(ExecutionVertexContext.ExecutionVertex.ResourceEntry)

PythonJobWorkerContext = _reflection.GeneratedProtocolMessageType('PythonJobWorkerContext', (_message.Message,), {
  'DESCRIPTOR' : _PYTHONJOBWORKERCONTEXT,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.PythonJobWorkerContext)
  })
_sym_db.RegisterMessage(PythonJobWorkerContext)

BoolResult = _reflection.GeneratedProtocolMessageType('BoolResult', (_message.Message,), {
  'DESCRIPTOR' : _BOOLRESULT,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.BoolResult)
  })
_sym_db.RegisterMessage(BoolResult)

Barrier = _reflection.GeneratedProtocolMessageType('Barrier', (_message.Message,), {
  'DESCRIPTOR' : _BARRIER,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.Barrier)
  })
_sym_db.RegisterMessage(Barrier)

CheckpointId = _reflection.GeneratedProtocolMessageType('CheckpointId', (_message.Message,), {
  'DESCRIPTOR' : _CHECKPOINTID,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.CheckpointId)
  })
_sym_db.RegisterMessage(CheckpointId)

BaseWorkerCmd = _reflection.GeneratedProtocolMessageType('BaseWorkerCmd', (_message.Message,), {
  'DESCRIPTOR' : _BASEWORKERCMD,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.BaseWorkerCmd)
  })
_sym_db.RegisterMessage(BaseWorkerCmd)

WorkerCommitReport = _reflection.GeneratedProtocolMessageType('WorkerCommitReport', (_message.Message,), {
  'DESCRIPTOR' : _WORKERCOMMITREPORT,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.WorkerCommitReport)
  })
_sym_db.RegisterMessage(WorkerCommitReport)

WorkerRollbackRequest = _reflection.GeneratedProtocolMessageType('WorkerRollbackRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORKERROLLBACKREQUEST,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.WorkerRollbackRequest)
  })
_sym_db.RegisterMessage(WorkerRollbackRequest)

CallResult = _reflection.GeneratedProtocolMessageType('CallResult', (_message.Message,), {
  'DESCRIPTOR' : _CALLRESULT,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.CallResult)
  })
_sym_db.RegisterMessage(CallResult)

QueueRecoverInfo = _reflection.GeneratedProtocolMessageType('QueueRecoverInfo', (_message.Message,), {

  'CreationStatusEntry' : _reflection.GeneratedProtocolMessageType('CreationStatusEntry', (_message.Message,), {
    'DESCRIPTOR' : _QUEUERECOVERINFO_CREATIONSTATUSENTRY,
    '__module__' : 'protobuf.remote_call_pb2'
    # @@protoc_insertion_point(class_scope:ray.streaming.proto.QueueRecoverInfo.CreationStatusEntry)
    })
  ,
  'DESCRIPTOR' : _QUEUERECOVERINFO,
  '__module__' : 'protobuf.remote_call_pb2'
  # @@protoc_insertion_point(class_scope:ray.streaming.proto.QueueRecoverInfo)
  })
_sym_db.RegisterMessage(QueueRecoverInfo)
_sym_db.RegisterMessage(QueueRecoverInfo.CreationStatusEntry)


DESCRIPTOR._options = None
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_CONFIGENTRY._options = None
_EXECUTIONVERTEXCONTEXT_EXECUTIONVERTEX_RESOURCEENTRY._options = None
_QUEUERECOVERINFO_CREATIONSTATUSENTRY._options = None
# @@protoc_insertion_point(module_scope)
